package com.hipme.flipsheet;

import danger.app.*;
import danger.audio.Meta;
import danger.internal.Date;
import danger.ui.*;
import danger.system.*;
import java.util.Vector;
import danger.util.DEBUG;

final class SheetScrollView extends ScrollView implements Constants {
  SheetTable sheetTable;
  int w, h;

  boolean tiled = true;
  boolean isTiled() { return tiled; }  
  
  SheetScrollView(int w, int h) {
    super();
    resize(w, h);
  }

  void resize (int w, int h) {
    this.w = w;
    this.h = h;
    setTiled(isTiled());
  }

  void setTiled(boolean tiled) {
    int myW = w;
    int myH = h;
    int myX = 0;
    int myY = 0;
    if (tiled) {
      myY += CellContentsTextField.FIELD_HEIGHT;
      myH -= CellContentsTextField.FIELD_HEIGHT;
    }
    setPosition(myX, myY);
    setSize(myW, myH);
    setOrigin(0, 0);
    setActualSize(myW, myH);
    this.tiled = tiled;
  }

  void setSheetTable(SheetTable sheetTable) {
    this.sheetTable = sheetTable;
    removeAllChildren();
    addChild(sheetTable);
    Font font = Font.findSystemFont();
    setScrollDelta(2 * 
		   (font.getAscent() + font.getDescent() + 
		    sheetTable.getBorderWidth() + sheetTable.getCellSpacing() + 
		    sheetTable.getCellPadding()));
  }

  void sendUpdateCell() {
    new Event(getParent(), kUpdateCell).sendToTarget();
  }

  void sendEditCellNoSelect() {
    Event e = new Event(getParent(), kEditCell);
    e.what = WHAT_NO_MOTION;
    e.sendToTarget();
  }

  void sendExitApp() {
    new Event(Application.getCurrentApp(), kReturnToLauncher).sendToTarget();
  }

  public boolean eventWidgetUp(int i, Event event) {
    boolean consumed = false;
    if (i == event.DEVICE_WHEEL_BUTTON) {
      sendEditCellNoSelect();
      consumed = true;
    } else if (i == event.DEVICE_WHEEL || i == event.DEVICE_MULTIPLE_WHEEL) {
      if (sheetTable.isWheelAxisHorizontal())
	wheelLeft(event.getWheelCount());
      else
	wheelUp(event.getWheelCount());
      consumed = true;
    } else {
      consumed = super.eventWidgetUp(i, event);
    }
    return consumed;
  }

  public boolean eventWidgetDown(int i, Event event) {
    boolean consumed = false; 
    if (i == event.DEVICE_WHEEL || i == event.DEVICE_MULTIPLE_WHEEL) {
      if (sheetTable.isWheelAxisHorizontal()) {
	wheelRight(event.getWheelCount());
      } else {
	wheelDown(event.getWheelCount());
      }
      consumed = true;
    } else if (i == event.DEVICE_ARROW_RIGHT) {
      sheetTable.nextCol();
      sendUpdateCell();
      consumed = true;
    } else if (i == event.DEVICE_ARROW_LEFT) {
      sheetTable.previousCol();
      sendUpdateCell();
      consumed = true;
    } else if (i == event.DEVICE_ARROW_DOWN) {
      arrowDown();
      consumed = true;
    } else if (i == event.DEVICE_ARROW_UP) {
      arrowUp();
      consumed = true;
    }
    else if (i == event.DEVICE_BUTTON_BACK) {
      // X exits app
      sendExitApp();
      consumed = true;
    } else {
      consumed = super.eventWidgetDown(i, event);
    }
    return consumed;
  }

  private void arrowUp() {
    sheetTable.previousRow();
    sendUpdateCell();
    keepCellInView();
  }

  private void arrowDown() {
    sheetTable.nextRow();
    sendUpdateCell();
    keepCellInView();
  }


  private void wheelUp(int rows) {
    for (int i = 0; i < rows; i++)
      sheetTable.previousRow();
    sendUpdateCell();
    keepCellInView();
  }

  private void wheelDown(int rows) {
    for (int i = 0; i < rows; i++)
      sheetTable.nextRow();
    sendUpdateCell();
    keepCellInView();
  }

  private void wheelLeft(int cells) {
    boolean rowChanged = false;
    for (int i = 0; i < cells; i++) {
      if (sheetTable.backCell()) 
	rowChanged = true;
    }
    sendUpdateCell();
    if (rowChanged)
      keepCellInView();
  }

  private void wheelRight(int cells) {
    boolean rowChanged = false;
    for (int i = 0; i < cells; i++) {
      if (sheetTable.forwardCell())
	rowChanged = true;
    }
    sendUpdateCell();
    if (rowChanged)
      keepCellInView();
  }

  private Rect rectToKeepInView = new Rect();  // Is this a good idea?

  void keepCellInView() {
    keepRectInView(sheetTable, sheetTable.getInterestingRect(rectToKeepInView));
  }
}
