package com.hipme.flipsheet;


// import danger.net.HiptopConnection;
import danger.app.*;
import danger.app.AppResources;
import danger.app.Application;
import danger.app.Event;
import danger.app.IPCMessage;
import danger.app.Registrar;
import danger.audio.Meta;
import danger.internal.Date;
import danger.system.*;
import danger.ui.*;
import danger.util.DEBUG;
import danger.util.format.StringFormat;
import java.util.Vector;

/**
 * Implements the main window class.
 *
 */
final class SheetWindow extends ScreenWindow implements Resources, Events, Constants {
  SheetScrollView sheetScrollView;
  SheetTable sheetTable;
  CellContentsTextField cellContentsTextField;
  EditText onLoadEditor;
  String []sheetNamesCache = null;

  public SheetWindow(String csv) {
    buildWorld(csv);
    updateCellContents();
  }
	
  public void Update(String csv) {		
    buildTable(csv);
    updateCellContents();
  }

  /**
   * Commit any outstanding edits before suspend, save, mail, etc.
   */
  void commitAnyEdits() {
    // crock -- but events are too late.
    if (cellContentsTextField != null && 
	cellContentsTextField == getFocusedChild()) {
      editCellContentsDone(cellContentsTextField.getStringCRLF());
    }
  }

  final String getCSV() {
    commitAnyEdits();
    return sheetTable.toCSV();
  }

  public boolean eventShortcut(char c, Event e) {
    boolean consumed = false;
    // eventShortcut currently is always null but it will be passed someday.
    // Put in true because this seems to fail.
    if (true || (e == null || e.modifiers == Event.EVENT_MODIFIER_MENU_BUTTON)) {
      switch (c) {
	// We only want Menu-Shift-T here, not Menu-t.
      case 'T': 
	toggleFullScreen();
	return true;
	// Menu-Shift-D invalidates dependency list.
      case 'D': 
	sheetTable.invalidateMasterDependencies();
	return true;
      }
    }
    return consumed || super.eventShortcut(c, e);
  }

  public boolean receiveEvent(Event e) {
    flipsheet app = ((flipsheet)Application.getCurrentApp());
    boolean consumed = false;
    switch (e.type) {
    case kEditCell:
      consumed = true;
      if (e.argument == null) {
	if (sheetTable.isCellLocked()) 
	  doMotion(e.what);	  
	else
	  editCellContents(e.what == WHAT_DOWN_OR_RIGHT_MOTION);
      } else {
	if (sheetTable.isCellLocked()) 
	  Meta.play(Meta.BEEP_COMMAND_REJECTED);
	else
	  editCellContents((e.what == WHAT_DOWN_OR_RIGHT_MOTION), ((String)e.argument));
      }
      break;
    case kUpdateCell:
      consumed = true;
      updateCellContents();
      break;
    case kEditCellDone:
      consumed = true;
      editCellContentsDone((String)(e.argument));
      if (e.what == WHAT_NO_MOTION) {
	updateCellContents();
      } else {
	doMotion(e.what);
	updateCellContents();
	sheetScrollView.keepCellInView();
      }
      break;
    case kAboutEvent:
      app.showAbout();
      consumed = true;
      break;
    case kHelpEvent:
      app.showHelp();
      consumed = true;
      break;
    case kDisplaySettingsEvent:
      app.showDisplaySettings();
      consumed = true;
      break;
    case kShowCellPropertiesEvent:
      app.showCellPropertiesDialog();
      consumed = true;
      break;
    case kToggleTiling:
      toggleTiled();
      consumed = true;
      break;
    case kImportFlipSheetEvent:
      ((flipsheet)Application.getCurrentApp()).showCSVEditor();
      invalidateSheetNames();
      consumed = true;
      break;
    case kConfirmClearSheetEvent:
      app.showDialog(kConfirmClearSheetDialog, this);
      consumed = true;
      break;
    case kClearSheetEvent:
      app.clearSheet();
      consumed = true;
      break;
    case kMailFlipSheetEvent:
      mailSheet(getCSV(), StringFormat.fromRsrc(kStrMailSheetFormat, getTitle()));
      consumed = true;
      break;
//    case kShowFlipFlipSheetEvent:
//      app.showTextInputAlert(kFlipFlipSheetTIA, this);
//      consumed = true;
//      break;
//    case kFlipFlipSheetEvent:
//      flipFlipSheet(getCSV(), TextInputAlertWindow.getTextFieldValue((IPCMessage)(e.argument), (kFlipFlipSheetTIAUserName)));
//      consumed = true;
//      break;
    case kEditOnLoad:
      app.showOnLoadEditor();
      consumed = true;
      break;
    case kSetFlipSheet:
      setSheetNumber(e.data);
      consumed = true;
      break;
    case kNewFlipSheet:
      setSheetNumber(e.data);
      new Event(this, kShowSheetPropertiesEvent).sendToTarget();
      invalidateSheetNames();
      consumed = true;
      break;
    case kShowSheetPropertiesEvent:
      app.showSheetPropertiesTIA();
      consumed = true;
      break;      
    case kCancelEdit:
      consumed = true;
      cancelEdit();
      break;
    default:
      consumed = super.receiveEvent(e);
    }
    return consumed;
  }

  void doMotion(int what) {
    switch(what) {
    case WHAT_DOWN_OR_RIGHT_MOTION:
      if (sheetTable.isWheelAxisHorizontal()) {
	do {
	  sheetTable.forwardCell();
	} while(sheetTable.isCellLocked() && sheetTable.getRowCursor() < sheetTable.nRows);
      } else {
	do {
	  sheetTable.nextRow();
	} while (sheetTable.isCellLocked() && sheetTable.getRowCursor() < sheetTable.nRows);
      }
      break;
    case WHAT_DOWN_MOTION:
      do {
	sheetTable.nextRow();
      } while (sheetTable.isCellLocked() && sheetTable.getRowCursor() < sheetTable.nRows);
      break;
    case WHAT_UP_MOTION:
      do {
	sheetTable.previousRow();
      } while (sheetTable.isCellLocked() && sheetTable.getRowCursor() >= 1);
      break;
    }
  }

  // If I don't capture the keyDown instgead of keyUp
  // then when I press enter
  // the textfield gets the key-up event and
  // exits focus on it.
  public boolean eventKeyUp(char c, Event e) {
    boolean consumed = false;
    if (cellContentsTextField != null && (getFocusedChild() == cellContentsTextField)) {
      consumed = super.eventKeyUp(c, e);
    } else if (c == 13) {
      editCellContents(true);
      consumed = true;
    } else {
      consumed = super.eventKeyUp(c, e);
    }
    return consumed;
  }

public boolean eventKeyDown(char c, Event e) {
  boolean consumed = false;
  if (cellContentsTextField != null && (getFocusedChild() == cellContentsTextField))
    return super.eventKeyDown(c, e);
  switch(c) {
  case ' ':
    sheetTable.setWheelAxisHorizontal(!sheetTable.isWheelAxisHorizontal());
    consumed = true;
    break;
  case '':
    editCellContents(false, "");		// should be done with a shortcut?
    consumed = true;
    break;
  }
  // Plain characters enter and start typing.
  if (!consumed && c > ' ' && c <= '~') {
    String s = sheetTable.getCellContents();
    if (s == null) s = "";
    // We have to send the whole event to the cell since modifier keys can be pressed and released.
    // Is there a better way to defer the event processing until after the textfield gets focus?
    editCellContents(false, s, c, e);
    consumed = true;
  }
  if (!consumed) {
    consumed = super.eventKeyUp(c, e);
  }
  return consumed;
}

  final void cancelEdit() {
    updateCellContents();
    setFocusedChild(sheetScrollView);
  }

  final void toggleTiled() {
    if (isTiled()) {
      sheetScrollView.setTiled(false);
      updateCellContents();
    } else {
      cellContentsTextField.show();
      cellContentsTextField.setEditable(false);
      sheetScrollView.setTiled(true);
      updateCellContents();
    }
  }

  final boolean isTiled() {
    return sheetScrollView.isTiled();
  }

  public void adjustActionMenuState() {
    ResourceDatabase resource = Application.getCurrentApp().getResources();
    Menu actionMenu = getActionMenu();
    actionMenu.removeAllItems();
    actionMenu.addFromResource(resource, kMainMenuResID, this);
    if (getFocusedChild() == cellContentsTextField) 
      actionMenu.addFromResource(resource, kMainMenuCancelResID, this);
    MenuItem item = actionMenu.addItem(resource.getString(kStrFlipsheetsMenu));
    Menu subMenu = resource.getMenu(kFlipsheetsSubMenu);
    adjustSheetMenuItems(subMenu);
    item.addSubMenu(subMenu);
  }

  void mailSheet(String csv, String subject) {
    IPCMessage ipc = new IPCMessage();
    ipc.addItem("action" , "send");
    if (subject != null) 
      ipc.addItem("subject" , subject);
    // it would be nice to do attachments instead
    ipc.addItem("body", csv);
    Registrar.sendMessage("Email" , ipc, null);
  }

//  int seqNo = 0;
//  void flipFlipSheet(String csv, String toUserName) {
//    DEBUG.p("flipping flipsheet to " + toUserName);
//    HiptopConnection hc = new HiptopConnection();
//    try {
//      hc.send_ThrowError(toUserName, csv.getBytes(), seqNo++);
//    } catch (java.io.IOException ioe) {
//      DEBUG.p("could not send to " + toUserName + ": " + ioe.getMessage());
//      ioe.printStackTrace();
//    }
//  }

  void setSheetNumber(int i) {
    flipsheet app = (flipsheet)Application.getCurrentApp();
    app.setSheetNumber(i);
  }

  void setSheetName(String name) {
    if (name == null) {
      DEBUG.p("ignoring setSheetName null!");
      return;
    }
    getSheetNamesCache()[((flipsheet)Application.getCurrentApp()).getSheetNumber()] = name;
    sheetTable.setSheetName(name);
    setTitle(name);
  }

  void invalidateSheetNames() {
    sheetNamesCache = null;
  }

  String [] getSheetNamesCache() {
    flipsheet app = (flipsheet)Application.getCurrentApp();
    int count = app.getSheetCount();
    if (sheetNamesCache == null || sheetNamesCache.length != count) {
      sheetNamesCache = new String[count];
      for (int i = 0; i < count; i ++) {
	try {
	  DataStore ds = app.getDataStore();
	  sheetNamesCache[i]=new CSVParser(app.getCSV(i)).sheetName;
	} catch (Exception e) {
	  DEBUG.p("Ignoring exception " + e + " trying to get name for sheet " + i);
	  e.printStackTrace();
	}
	if (sheetNamesCache[i] == null)
	  sheetNamesCache[i] = sheetTable.getDefaultSheetName(i); // crock
      }
    }
    return sheetNamesCache;
  }

  private void adjustSheetMenuItems(Menu menu) {
    flipsheet app = (flipsheet)Application.getCurrentApp();
    String[] names = getSheetNamesCache();
    int count = sheetNamesCache.length;
    for (int i = 0; i < count; i++) {
      menu.addItem(names[i],
		   kSetFlipSheet,
		   i,
		   null,
		   this).setShortcut((char)('0'+i));
    }
    // put in "New Sheet" item if not at max.
    if (count < flipsheet.MAX_SHEETS) {
      menu.addItem(app.getString(kStrNewFlipsheet), 
		   kNewFlipSheet,
		   count,
		   null,
		   this).setShortcut('n');
    }
    MenuItem current = getSheetMenuItem(menu, app.getSheetNumber());
    if (current != null) current.setChecked(true);
  }

  private MenuItem getSheetMenuItem(Menu menu, int i) {
    if (i < 0 || i >= flipsheet.MAX_SHEETS) return null;
    return menu.getItemFromShortcut((char)('0' + i));
  }

  final void editCellContents(boolean selectAll) {
    if (sheetTable.isCellLocked()) {
      if (selectAll)
	doMotion(WHAT_DOWN_OR_RIGHT_MOTION);	  
      else
	Meta.play(Meta.BEEP_COMMAND_REJECTED);
    }
    else
      editCellContents(selectAll, sheetTable.getCellContents());
  }


  final void editCellContents(boolean selectAll, String value) {
    if (sheetTable.isCellLocked()) {
      Meta.play(Meta.BEEP_COMMAND_REJECTED);
      return;
    }
    cellContentsTextField.editCellContents(selectAll, value);
    setSubTitle(sheetTable.getCellName());
    invalidate();
  }

  /**
   * Event e is the read-ahead event that should go to the text field.
   */
  final void editCellContents(boolean selectAll, String value, char c, Event e) {
    if (sheetTable.isCellLocked()) {
      Meta.play(Meta.BEEP_COMMAND_REJECTED);
      return;
    }
    cellContentsTextField.editCellContents(selectAll, value);
    // crock to redirect the down and up events to the TextField now that we've captured the
    // up event.
    Event down = new Event(e);
    down.setListener(cellContentsTextField);
    down.type = Event.EVENT_KEY_DOWN;
    Event up = new Event(down);
    up.type = Event.EVENT_KEY_UP;
    cellContentsTextField.eventKeyDown(c, down);
    cellContentsTextField.eventKeyUp(c, up);
    setSubTitle(sheetTable.getCellName());
    invalidate();
  }

  final void updateCellContents() {
    String value = sheetTable.getCellContents();
    cellContentsTextField.setContents(value);
    if (! isTiled())
      cellContentsTextField.hide();
    if (sheetTable.isCellLocked()) {
      // so I'd like use use Unicode 29BB but it fails
      // so I'll use this X mark.  Crock!
      setSubTitle(sheetTable.getCellName() + " " + danger.ui.Font.GLYPH_BACK);
    } else {
      setSubTitle(sheetTable.getCellName());
    }
  }

  final void editCellContentsDone(String value) {
    setFocusedChild(sheetScrollView); // this scrolls to sheetTable.getInterestingRect()
    String oldValue = sheetTable.getCellContents();
    cellContentsTextField.setEditable(false);
    if (! value.equals(oldValue)) {
      sheetTable.setCellContents(value);
      sheetTable.calculate();
    }
  }


  final void buildWorld(String csv) {
    setIcon(Application.getCurrentApp().getBitmap(AppResources.ID_SMALL_ICON));
    sheetScrollView = new SheetScrollView(getWidth(), getHeight());
    addChild(sheetScrollView);
    cellContentsTextField = new CellContentsTextField(this);
    cellContentsTextField.setEditable(false);
    addChild(cellContentsTextField);
    cellContentsTextField.show();
    buildTable(csv);
    sheetScrollView.show();
    if (getFocusedChild() == null) {
      setFocusedChild(sheetScrollView);
    }
  }

  final void buildTable(String csv) {
    sheetTable = new SheetTable(csv);
    sheetScrollView.setSheetTable(sheetTable);
    sheetTable.show();
  }

  public String GetWindowTitle() {
    return (Application.getCurrentApp().getString(kStrSheetWindowTitle));
  }

  final void toggleFullScreen() {
    setFullScreen(!isFullScreen());
    sheetScrollView.resize(getWidth(), getHeight());
    cellContentsTextField.resize();
    invalidate();
  }

  void clearSheet() {
    sheetTable.clear();
    invalidate();
  }
}
