// Copyright 2003 Leigh L. Klotz, Jr.

package com.hipme.flipsheet;

// import danger.net.HiptopTransaction;
import danger.app.*;
import danger.app.SettingsDB;
import danger.audio.Meta;
import danger.audio.Meta;
import danger.internal.Date;
import danger.net.HiptopConnection;
import danger.net.URL;
import danger.net.URLException;
import danger.system.*;
import danger.ui.*;
import danger.ui.AlertWindow;
import danger.ui.CheckBox;
import danger.ui.Color;
import danger.ui.DialogWindow;
import danger.ui.TextField;
import danger.ui.TextInputAlertWindow;
import danger.util.DEBUG;
import danger.util.format.StringFormat;
import java.util.Vector;

/**
 * Implements the main application class.
 *
 */
public final class flipsheet extends Application implements Resources, Events, Constants
{
  final static String CSV_DATASTORE_NAME = "CSV";
  final static int MAX_SHEETS = 10;

  private static SheetWindow sheetWindow;
  int sheetNumber = 0;
  int sheetUID = -1;

  final static String SETTINGS_DB="flipsheet";
  SettingsDB settings;
  final static String SHOW_DESCENDERS = "showDescenders";
  boolean showDescenders = false;
  final static String USE_OUTLINE = "useOutline";
  boolean useOutline = !Screen.isColor();

  /**
   * Creates the main application class.
   *
   */
  public flipsheet() {
    super();

    // init data store
    DataStore ds = getDataStore();
    if (ds == null) {
      DEBUG.p("flipsheet: no datastore found");
      ds = DataStore.createDataStore(CSV_DATASTORE_NAME, true);
      DEBUG.p("flipsheet: ds = " +ds);
      int rn = ds.addRecord(getString(kDefaultInitialFlipsheetCSV).getBytes());
    }
    ds.setAutoSyncNotifyee(this);
    DataStore.registerListenerForAutoSyncDone(this);

    // init settings
    boolean oldSettings = SettingsDB.findDB(SETTINGS_DB);
    settings = new SettingsDB(SETTINGS_DB, true);
    if (oldSettings) {
      restoreSettings();
    } else {
      initSettings();
    }
    // Registrar.registerURLSchemeHandler("flipsheet:");
  }


  void initSettings() {
    settings.setIntValue(SHOW_DESCENDERS, showDescenders ? 1 : 0);
    settings.setIntValue(USE_OUTLINE, useOutline ? 1 : 0);
  }

  void restoreSettings() {
    try {
      showDescenders = (settings.getIntValue(SHOW_DESCENDERS) != 0);
    } 
    catch (SettingsDBException settingsdbexception) { }
    try {
      useOutline = (settings.getIntValue(USE_OUTLINE) != 0);
    } 
    catch (SettingsDBException settingsdbexception) {
      useOutline = !Screen.isColor(); 
    }
  }
	
  public SheetWindow getMainWindow() {
    if (sheetWindow == null) {
      sheetWindow = new SheetWindow(null);
      sheetWindow.setTitle(getString(kStrSheetWindowTitle));
    }
    return sheetWindow;
  }

  /**
   * Handles the Suspend event.
   */
  public void suspend() {
    saveSheet();
    removeEmptySheets();
    super.suspend();
  }


  public boolean receiveEvent(Event e) {
    boolean consumed = false;
    switch (e.type) {
    case kReturnToLauncher:
    case kLockedBundleExit:
      consumed = true;
      returnToLauncher();
      break;
      //    Does not seem to happen.
    case Event.EVENT_DATASTORE_RESTORED:
      DEBUG.p("flipsheet: EVENT_DATASTORE_RESTORED");
      consumed = super.receiveEvent(e);
      break;
    case Event.EVENT_SETTINGSDB_FLUSHED:
      DEBUG.p("flipsheet: EVENT_SETTINGSDB_FLUSHED");
      consumed = super.receiveEvent(e);
      break;
    case Event.EVENT_HINGE_CHANGED:
      // This seems not to happen
      e.debugPrint();
      if (sheetWindow != null && getSheetTable() != null) {
	if (e.what == Event.WHAT_HINGE_CLOSED) {
	  getSheetTable().setWheelAxisHorizontal(true);
	} else {
	  getSheetTable().setWheelAxisHorizontal(false);
	}
      }
      consumed = super.receiveEvent(e);
      break;
    case kSetOnLoadEvent:
      onSetOnLoad((DialogWindow)(e.argument));
      consumed = true;
      break;
    case kCSVEditDoneEvent:
      onCSVEditDone((DialogWindow)(e.argument));
      consumed = true;
      break;
    case kSheetPropertiesEvent:
      onSheetProperties((IPCMessage)(e.argument));
      consumed = true;
      break;      
    case kCellPropertiesEvent:
      onCellProperties((DialogWindow)(e.argument));
      consumed = true;
      break;
    case kShowLockedBundleAlert:
      showLockedBundleAlert();
      consumed = true;
      break;
    case kDisplaySettingsEvent:
      showDisplaySettings();
      consumed = true;
      break;
    case kDisplaySettingsDoneEvent:
      handleDisplaySettingsDoneEvent((DialogWindow)(e.argument));
      consumed = true;
      break;
    default:
      consumed = super.receiveEvent(e);
    }
    return consumed;
  }


  private void showLockedBundleAlert() {
    AlertWindow a = getAlert(kLockedBundleAlert, this);
    a.setMessage(StringFormat.fromRsrc(kLockedBundleAlertText, getString(kLockedBundleUser)));
    a.show();
  }

  /**
   * Handles the Resume event.
   *
   */
  public void resume() {
    System.err.println("flipsheet: Version = " + getString(ID_VERSION));
    DataStore ds = getDataStore();
    // See if records got reordered.
    boolean mismatch = (ds != null) && (ds.getRecordUID(sheetNumber) != sheetUID);
    if ((sheetWindow != null) && mismatch)
      sheetWindow.invalidateSheetNames();
    // Create/get the main window...
    if (sheetWindow == null || mismatch) {
      sheetWindow = getMainWindow();
      // Get the data store and read the CSV
      String csv = getCSV();
      // Show the main window...
      sheetWindow.Update(csv);
    }

    sheetWindow.setTitle(getSheetTable().getSheetName());
    sheetWindow.show();

    // check for locked bundle
    String user = (getString(kLockedBundleUser));
    if (user.length()>0 && !HiptopConnection.getUserName().startsWith(user)) {
      new Event(this, kShowLockedBundleAlert).sendToTarget();
    }
  }
	
  DataStore getDataStore() {
    return DataStore.findDataStore(CSV_DATASTORE_NAME);
  }


  /**
   * Commit edits, and save sheet as CSV, even if empty.
   */
  void saveSheet() {
    putCSV(sheetWindow.getCSV());
  }

  /**
   * On suspend, flush any empty sheets.  We don't do this while the app is running
   * to avoid re-ordering the display when someone creates a new sheet and navigates
   * off it.
   * Remove any empty sheets (except 0 and current sheet).
   * Updates sheetNumber without redisplay so call only on suspend().
   */
  private void removeEmptySheets() {
    DataStore ds = getDataStore();
    if (ds != null) {
      for (int i = 1; i < getSheetCount(); i++) {
	if (i != sheetNumber) {
	  // there should be a faster way of tell what was empty in this session.
	  if (ds.getRecordData(i).length == 0) {
	    DEBUG.p("flipsheet: Dropping empty record " + i);
	    ds.removeRecord(i);
	    // move sheet number down behind its back.
	    if (i < sheetNumber) sheetNumber--;
	  }
	}
      }
    }
  }

  void putCSV(String csv) {
    DataStore ds = getDataStore();
    DEBUG.p("flipsheet: putting CSV to " + ds);
    if (ds == null) throw new RuntimeException("Cannot happen -- no DataStore " + CSV_DATASTORE_NAME);
    if (ds.getRecordCount() <= sheetNumber) {
      int rn = ds.addRecord(csv.getBytes());
    } else {
      ds.setRecordData(sheetNumber, csv.getBytes());
    }
    ds.discardOnReboot(false);	// should have done this when making it.
    DEBUG.p("flipsheet: Record " + sheetNumber + " set: total bytes = " + ds.getBytesUsed());
  }

  /**
   * Returns CSV string or null if none.
   */

  String getCSV() {
    return getCSV(sheetNumber);
  }

  String getCSV(int sn) {
    String result = null;
    DataStore ds = getDataStore();
    byte[] rec = ds.getRecordData(sn);
    // We get no records if we ask for data before the connection is up. 
    // How do we prevent this?
    if (rec != null) {
      result = new String(rec);
    } else {
      DEBUG.p("flipsheet: Got empty CSV result for " + sn);
    }
    return result;
  }

  void setSheetNumber(int i) {
    if (i == sheetNumber) return;
    if (i < 0 || i >= MAX_SHEETS) {
      DEBUG.p("flipsheet: sheet number out of range: " + i);
      Meta.play(Meta.BEEP_ACTION_FAIL);
      return;
    }
    saveSheet();
    sheetNumber = i;
    sheetWindow.Update(getCSV());
    if (sheetNumber >= getSheetCount()) {
      saveSheet();
    }
    DataStore ds = getDataStore();
    if (ds != null) 
      sheetUID = ds.getRecordUID(sheetNumber);
    sheetWindow.setTitle(getSheetTable().getSheetName());
    Meta.play(Meta.BEEP_ACTION_SUCCESS);    
  }

  int getSheetNumber() {
    return sheetNumber;
  }

  int getSheetUID() {
    return sheetUID;
  }

  int getSheetCount() {
    DataStore ds = getDataStore();
    if (ds == null) return 0;
    return ds.getRecordCount();
  }

  void showAbout() {
    AlertWindow a = getAlert(kAboutAlert, this);
    String user = getString(kLockedBundleUser);
    String v = (user.length()==0) ? getString(ID_VERSION) : getString(ID_VERSION) + "-"+ user;
    a.setMessage(StringFormat.fromRsrc(kAboutAlertText, v));
    a.show();
  }

  void showHelp() {
    try {
      URL.gotoURL(getString(kHelpURL) + getString(ID_VERSION));
    }
    catch (URLException e) {
      DEBUG.p("flipsheet: error getting help" + e);
      e.printStackTrace();
    }
  }

  // kDisplaySettingsEvent
  void showDisplaySettings() {
    DialogWindow settingsDialog = getDialog(kDisplaySettingsDialog, this);
    boolean compact = !showDescenders;
    ((CheckBox)(settingsDialog.getDescendantWithID(kDisplaySettingsCompactCheckBox))).setValue(compact ? 1 : 0);
    ((CheckBox)(settingsDialog.getDescendantWithID(kDisplaySettingsUseOutlineCheckBox))).setValue(useOutline ? 1 : 0);
    settingsDialog.show();
  }

  // kDisplaySettingsDoneEvent
  void handleDisplaySettingsDoneEvent(DialogWindow settingsDialog) {
    SheetTable sheetTable = getSheetTable();
    {
      int compact = (((CheckBox)(settingsDialog.getDescendantWithID(kDisplaySettingsCompactCheckBox))).getValue());
      showDescenders = (compact == 0);
      settings.setIntValue(SHOW_DESCENDERS, showDescenders ? 1 : 0);
      sheetTable.setShowDescenders(showDescenders);
    }
    {
      int val = (((CheckBox)(settingsDialog.getDescendantWithID(kDisplaySettingsUseOutlineCheckBox))).getValue());
      useOutline = (val != 0);
      settings.setIntValue(USE_OUTLINE, useOutline ? 1 : 0);
      sheetTable.setUseOutline(useOutline);
    }
  }

  void showCircularDependency(String message) {
    AlertWindow a = getAlert(kCircularDependencyAlert, this);
    a.setMessage(message);
    a.show();
    Meta.play(Meta.BEEP_ACTION_FAIL);
  }

  void showOnLoadEditor() {
    DialogWindow onLoadEditor = getDialog(kEditOnLoadDialog, this);
    EditText t = ((EditText)(onLoadEditor.getDescendantWithID(kID_EditOnLoadText)));
    t.setText(getSheetTable().getOnLoadText());
    onLoadEditor.show();
    onLoadEditor.setFocusedChild(t); // why do I have to do this?
  }

  void onSetOnLoad(DialogWindow onLoadEditor) {
    EditText t = ((EditText)(onLoadEditor.getDescendantWithID(kID_EditOnLoadText)));
    if (t == null) return;
    getSheetTable().setOnLoadText(t.toString());
  }

  void showCSVEditor() {
    String csv = sheetWindow.getCSV();
    if (csv == null)
      csv = "";
    DialogWindow editCSVEditor = getDialog(kEditCSVDialog, this);
    EditText t = ((EditText)(editCSVEditor.getDescendantWithID(kID_EditCSVText)));
    t.setText(csv);
    t.selectAll();
    editCSVEditor.show();
    editCSVEditor.setFocusedChild(t); // why do I have to do this?
  }

  void onCSVEditDone(DialogWindow editCSVEditor) {
    EditText t = ((EditText)(editCSVEditor.getDescendantWithID(kID_EditCSVText)));
    if (t == null) return;
    sheetWindow.Update(t.toString());
    saveSheet();
  }

  void showSheetPropertiesTIA() {
    SheetTable sheetTable = getSheetTable();
    TextInputAlertWindow tia = getTextInputAlert(kSheetPropertiesTIA, this);
    {
      TextField t = ((TextField)(tia.getTextFieldWithID(kSheetPropertiesTIAName)));
      t.setText(sheetTable.getSheetName());
    }
    {
      TextField t = ((TextField)(tia.getTextFieldWithID(kSheetPropertiesTIAPrecision)));
      t.setText(String.valueOf(sheetTable.getPrecision()));
    }
    tia.show();
  }

  void onSheetProperties(IPCMessage message) {
    SheetTable sheetTable = getSheetTable();
    String name = TextInputAlertWindow.getTextFieldValue(message, kSheetPropertiesTIAName);
    String precision = TextInputAlertWindow.getTextFieldValue(message, kSheetPropertiesTIAPrecision);
    if (name.length()==0) return;
    if (precision.length()==0) precision="0";
    sheetWindow.setSheetName(name);
    sheetTable.setPrecision(Integer.valueOf(precision).intValue());
    // a bit harsh but ok for now
    sheetTable.calculate();
  }

  void showCellPropertiesDialog() {
    DialogWindow dialog = getDialog(kCellPropertiesDialog, this);
    SheetTable sheetTable = getSheetTable();
    {
      CheckBox t = ((CheckBox)(dialog.getDescendantWithID(kCellPropertiesDialogLocked)));
      t.setValue(sheetTable.isCellLocked() ? 1 : 0);
    }
    {
      TextField t = ((TextField)(dialog.getDescendantWithID(kCellPropertiesDialogFG)));
      t.setText(colortoString(((SheetTableCell)(sheetTable.getCell())).foregroundColor));
    }
    {
      TextField t = ((TextField)(dialog.getDescendantWithID(kCellPropertiesDialogBG)));
      t.setText(colortoString(((SheetTableCell)(sheetTable.getCell())).backgroundColor));
    }
    dialog.show();
  }

  String colortoString(int color) {
    switch(color & 0xffffff) {
    case Color.BLACK:
	return "black";
    case Color.WHITE:
	return "white";
    case Color.RED:
	return "red";
    case Color.GREEN:
	return "green";
    case Color.BLUE:
	return "blue";
    case Color.CYAN:
	return "cyan";
    case Color.MAGENTA:
	return "magenta";
    case Color.ORANGE:
	return "orange";
    case Color.PURPLE:
	return "purple";
    case Color.VIOLET:
	return "violet";
    case Color.YELLOW:
	return "yellow";
    case Color.PINK:
	return "pink";
    case Color.GRAY1:
	return "gray1";
    case Color.GRAY2:
	return "gray2";
    case Color.GRAY3:
	return "gray3";
    case Color.GRAY4:
	return "gray4";
    case Color.GRAY5:
	return "gray5";
    case Color.GRAY6:
	return "gray6";
    case Color.GRAY7:
	return "gray7";
    case Color.GRAY8:
	return "gray8";
    case Color.GRAY9:
	return "gray9";
    case Color.GRAY10:
	return "gray10";
    case Color.GRAY11:
	return "gray11";
    case Color.GRAY12:
	return "gray12";
    case Color.GRAY13:
	return "gray13";
    case Color.GRAY14:
	return "gray14";
    default:
      return '#' + Color.toString(color);
    }
  }

  /**
   * When confirming cell properties, if any have changed,
   * add them to the Evaluate On Load list and eval them then.
   * Some day addOnLoadText will eliminate dups and flip/flops.
   */
  void onCellProperties(DialogWindow dialog) {
    SheetTable sheetTable = getSheetTable();
    String cellName = sheetTable.getCellName();
    {
      CheckBox t = ((CheckBox)(dialog.getDescendantWithID(kCellPropertiesDialogLocked)));
      boolean lock = (t.getValue() != 0);
      if (sheetTable.isCellLocked() != lock) {
	sheetTable.addOnLoadText("lock("+t.getValue()+", " + cellName +")");
      }
    }
    {
      TextField t = ((TextField)(dialog.getDescendantWithID(kCellPropertiesDialogFG)));
      String fg = t.toString();
      if (!Color.colorsEqual(((SheetTableCell)(sheetTable.getCell())).foregroundColor, Color.fromString(fg))) {
	// hack: setcolor understands #<symbol green> in addition to String "green",
	// which means we don't have to figure out if this is a number or a #hex number or a color name.
	sheetTable.addOnLoadText("setcolor("+fg+", " + cellName +")");
      }
    }
    {
      TextField t = ((TextField)(dialog.getDescendantWithID(kCellPropertiesDialogBG)));
      String bg = t.toString();
      if (!Color.colorsEqual(((SheetTableCell)(sheetTable.getCell())).backgroundColor, Color.fromString(bg))) {
	sheetTable.addOnLoadText("setbg("+bg+", " + cellName+")");
      }
    }
  }

  public void clearSheet() {
    DEBUG.p("flipsheet: clearing sheet " + sheetNumber);
    sheetWindow.clearSheet();
  }

  final SheetTable getSheetTable() {
    return sheetWindow.sheetTable;
  }

//  public final void networkEvent(Object item) {
//    DEBUG.p("networkEvent " + item);
//    if (item instanceof HiptopTransaction) {
//      HiptopTransaction ht = ((HiptopTransaction)item);
//      String u = "UNKNOWN USER"; // ht.getUserName();
//      DEBUG.p("received flipsheet from " + u);
//      setSheetNumber(sheetNumber+1);
//      Meta.play(Meta.BEEP_ACTION_SUCCESS);
//      sheetWindow.Update(ht.getString());
//    } else {
//      super.networkEvent(item);
//    }
//  }
}
