package com.hipme.flipsheet;

import danger.app.*;
import danger.ui.*;
import danger.util.*;


final class CellContentsTextField extends TextField implements Events, Resources, Constants {
  // How can I calculate this?
  static final int FIELD_HEIGHT = 17;
  // How can I find this out?
  static final int BORDER_WIDTH = 3;
  Container owner;
  int width;

  CellContentsTextField(Container owner) {
    this.owner=owner;
    setAllowCursorBlink(true);
    setAutoText(false);
    setAutoCap(true);
    // Should this be done in Layout?
    setFont(Font.findBoldSystemFont());
    width = owner.getWidth();
    resize();
  }
	
  void resize() {
    if (getHasBorder()) 
      setPosition(0, 0);
    else 
      setPosition(BORDER_WIDTH, BORDER_WIDTH);
    setSize(width, FIELD_HEIGHT);
  }

  void setContents(String value) {
    setText(value);
    invalidate(getBounds());
  }

  void sendEditDone(int moveAfter) {
    String value = getStringCRLF();
    Event doneEvent = new Event(getParent(), kEditCellDone);
    doneEvent.setArgument(value);
    doneEvent.what = moveAfter;
    doneEvent.sendToTarget();
  }

  public boolean receiveEvent(Event e) {
    boolean consumed = true;
    switch (e.type) {
    case kEditCellCancel:
      cancelEdit();
      break;
    default:
      consumed = super.receiveEvent(e);
      break;
    }
    return consumed;
  }

  private void cancelEdit() {
    SheetWindow p = (SheetWindow)getParent();
    p.cancelEdit();
  }

  void editCellContents(boolean selectAll, String value) {
    setContents(value);
    setEditable(true);
    show();
    ((SheetWindow)getParent()).setFocusedChild(this);
    if (selectAll)
      selectAll();
    else
      setInsertionPoint(length());
  }

  public void setEditable(boolean editable) {
    super.setEditable(editable);
    setHasBorder(editable);
    resize();
  }

  public void loseFocus() {
    // Stop being editable when we lose focus.
    if (isEditable())
      setEditable(false);
    super.loseFocus();
  }

  public boolean eventWidgetDown(int i, Event event) {
    boolean consumed = false;
    if (i == event.DEVICE_BUTTON_BACK) {
      cancelEdit();
      consumed = true;
    } else if (i == event.DEVICE_ARROW_DOWN) {
      sendEditDone(WHAT_DOWN_MOTION);
    } else if (i == event.DEVICE_ARROW_UP) {
      sendEditDone(WHAT_UP_MOTION);
    } 
    else if (i == event.DEVICE_ARROW_LEFT || i == event.DEVICE_ARROW_RIGHT) {
      // It's too easy to scroll left/right off the edit field,
      // so we don't make left/right commit.
      // Also, in TextField, when point is at bol,
      // left arrow returns false to pass
      // the event to parent,
      // which shifts focus to the ScrollView.
      // That's not what we want, so we
      // always consume the event here.
      super.eventWidgetDown(i, event);
      consumed = true;
    } else {
      consumed = super.eventWidgetDown(i, event);
    }
    return consumed;
  }

  public boolean eventWidgetUp(int i, Event event) {
    boolean consumed = false;
    if (i == event.DEVICE_WHEEL_BUTTON) {
      sendEditDone(WHAT_NO_MOTION);
      consumed = true;
    } else {
      consumed = super.eventWidgetUp(i, event);
    }
    return consumed;
  }

  public boolean eventKeyUp(char c, Event event) {
    boolean consumed = false;
    if(c == 13) {
      sendEditDone(WHAT_DOWN_OR_RIGHT_MOTION);
      consumed = true;
    } else {
      consumed = super.eventKeyUp(c, event);
    }
    return consumed;
  }
}
