package com.hipme.flipsheet;

import danger.app.Event;
import danger.app.EventType;
import danger.ui.*;
import danger.audio.Meta;
import danger.util.DEBUG;


/** This class should not have to exist, 
 * but there are bugs in ScrollView in hiptop OS 1.1
 * that make horizontal offsets fail.
 * This class reproduces enough of danger.ui.ScrollView's functionality
 * that com.hipme.SheetScrollView can specialize it, and still be
 * forward compatible when Danger fixes ScrollView.
 * 
 * Probably some of it can be commented out, since SheetScrollView
 * uses it with only one child, and so the focusing can be removed.
 */
public class MyScrollView extends Container implements EventType {

  int mScrollBottom;
  int mScrollDelta;
  Rect mWorkRect1;
  Rect mWorkRect2;
  Point mWorkPoint;
  private boolean mHasScrollbar;
  private Scrollbar mScrollbar;
  int mScrollContext = 20;
  int mHorizontalScrollContext = 0;

  public void reflowChildren(View view, int i)
  {
    super.reflowChildren(view, i);
    View view3 = getBottomChild();
    Rect rect4;
    Point point;
    Rect rect;
    if (view3 == null) 
      return;
         
    mScrollBottom = view3.getFrame(mWorkRect1).bottom;
    if (i < 0) {
      rect = getBounds(mWorkRect1);
      rect4 = view3.getFrame(mWorkRect2);
      point = getOrigin(mWorkPoint);
      if (rect4.bottom + point.v < rect.bottom) 
	setOrigin(rect.bottom - rect4.bottom);
              
    }
    else if (getFocusedChild() == view) {
      rect = view.getInterestingRect();
      point = getOrigin(mWorkPoint);
      rect.translate(point);
      rect.top = rect.top + view.getTop();
      rect.bottom = rect.bottom + view.getTop();
      rect4 = getBounds(mWorkRect2);
      if (rect.top <= rect4.top) 
	setOrigin(rect4.top - rect.top);
              
      else if (rect.bottom > rect4.bottom) 
	setOrigin(rect4.bottom - (rect.bottom + i));
              
    }
    getParent().invalidate();
    return;
  }


  public void addChild(View view ,int i)
  {
    super.addChild(view, i);
    mScrollBottom = getBottomChild().getFrame().bottom;
    return;
  }


  public void calcBottom()
  {
    View view = getBottomChild();
    if (view != null) 
      mScrollBottom = getBottomChild().getFrame().bottom;
         
    return;
  }


  public void removeAllChildren()
  {
    super.removeAllChildren();
    mScrollBottom = 0;
    setOrigin(0);
    return;
  }


  public void setSize(int i ,int j)
  {
    return;
  }


  public void setActualSize(int i ,int j)
  {
    super.setSize(i, j);
    return;
  }


  public void setActualWidth(int i)
  {
    setActualSize(i, getHeight());
    return;
  }


  public void setActualHeight(int i)
  {
    setActualSize(getWidth(), i);
    return;
  }


  public boolean eventWidgetDown(int i ,Event event)
  {
    int j = 0;
    boolean consumed = false;
    if (getFocusedChild() != null && getFocusedChild().eventWidgetDown(i, event)) 
      return (true);
         
    switch (i) {
              
    case DEVICE_ARROW_DOWN: 
    case DEVICE_ARROW_RIGHT: 
    case DEVICE_WHEEL: 
      consumed = advanceFocusOrScroll(event.getWheelCount());
      break;
                            
    case DEVICE_MULTIPLE_WHEEL: 
      consumed = advanceFocusOrScroll(event.getWheelCount());
      break;
                            
    case DEVICE_ARROW_UP: 
    case DEVICE_ARROW_LEFT: 
      consumed = reverseFocusOrScroll(event.getWheelCount());
      break;
                            
    case DEVICE_WHEEL_PAGE_DOWN: 
      consumed = pageDown();
                       
    }
    return (consumed);
  }


  public boolean eventWidgetUp(int i ,Event event)
  {
    int j = 0;
    boolean consumed = false;
    if (getFocusedChild() != null && getFocusedChild().eventWidgetUp(i, event)) 
      return (true);
         
    switch (i) {
              
    case DEVICE_WHEEL: 
      consumed = reverseFocusOrScroll(event.getWheelCount());
      break;
                            
    case DEVICE_MULTIPLE_WHEEL: 
      consumed = reverseFocusOrScroll(event.getWheelCount());
      break;
                            
    case DEVICE_WHEEL_PAGE_UP: 
      consumed = pageUp();
                       
    }
    return (consumed);
  }


  protected boolean focusChangedTo(View view ,boolean i)
  {
    if (canChangeFocusTo(view, i)) {
      setFocusedChild(view, i);
      return (true);
    }
    return (false);
  }


  protected boolean canChangeFocusTo(View view ,boolean i)
  {
    Point point;
    Rect rect;
    if (view == null) 
      return (false);
         
    point = getOrigin(mWorkPoint);
    rect = getBounds(mWorkRect1);
    rect.translate(-point.h, -point.v);
    if (view.willFocusWithin(rect, i)) 
      return (true);
         
    return (false);
  }


  protected boolean advanceFocusOrScroll(int i)
  {
    if (tryToAdvanceFocus()) 
      return (true);
         
    if (getHeight() > mScrollBottom || getHeight() - mScrollBottom == mOriginY) {
      playLimitSound(true);
      return (false);
    }
    offsetOrigin(-mScrollDelta * i);
    getParent().invalidate();
    return (true);
  }


  private final boolean tryToAdvanceFocus()
  {
    View view = getNextFocusableChild();
    if (canChangeFocusTo(view, true)) {
      focusChangedTo(view, true);
      return (true);
    }
    for (;view != null;) {
      if (canChangeFocusTo(view, true)) {
	focusChangedTo(view, true);
	return (true);
      }
      view = getNextFocusableChild(view);
    }
    return (false);
  }


  public void offsetOriginAndSelect(int i)
  {
    int j = mOriginY;
    offsetOrigin(i);
    boolean k;
    View view;
    int o;
    if (i > 0) {
      if (mOriginY - j >= i) 
	k = false;
              
      else
	k = true;
              
    }
    else if (mOriginY - j <= i) 
      k = true;
         
    else
      k = false;
         
    o = 0;
    for (;o < 3;) {
      if (k)
	view = MYgetFirstFocusableChild(this);
              
      else
	view = MYgetLastFocusableChild(this);
              
      for (;view != null;) {
	if (o != 0 && entirelyVisible(view) || o != 1 && atAllVisible(view) || o == 2) {
	  if (view == getFocusedChild()) 
	    if (!(view instanceof Container) && tryToFocusWithin((Container)view, k)) 
	      return;
                             
                        
	    else if (canChangeFocusTo(view, k)) {
	      if (view instanceof Container) 
		if (o == 2 && !tryToFocusWithin((Container)view, k) && focusChangedTo(view, k)) 
		  return;
                                  
                             
		else if (focusChangedTo(view, k)) 
		  return;
                             
	    }
	}
	if (k) {
	  view = getNextFocusableChild(view);
	  continue;
	}
	view = getPrevFocusableChild(view);
      }
      o++;
    }
    return;
  }


  private final boolean tryToFocusWithin(Container container ,boolean i)
  {
    Point point = getOrigin(mWorkPoint);
    Rect rect = getBounds(mWorkRect1);
    rect.translate(-point.h, -point.v);
    return (tryToFocusWithin(container, i, rect.left, rect.top, rect.right, rect.bottom));
  }


  private final boolean tryToFocusWithin(Container container ,boolean i ,int j ,int k ,int l ,int m)
  {
    int n = 0;
    for (;n < 3;) {
      int q;
      Rect rect;
      int o;
      int r;
      Point point;
      View view;
      int p;
      if (i) 
	view = MYgetFirstFocusableChild(container);
              
      else
	view = MYgetLastFocusableChild(container);
              
      if (view == null) 
	return (true);
              
      for (;view != null;) {
	rect = mWorkRect1;
	rect.left = j;
	rect.top = k;
	rect.right = l;
	rect.bottom = m;
	point = container.getOrigin(mWorkPoint);
	rect.translate(-point.h, -point.v);
	rect.translate(-container.getLeft(), -container.getTop());
	o = rect.left;
	p = rect.top;
	q = rect.right;
	r = rect.bottom;
	if (n != 0 && entirelyVisible(view, rect) && n != 1 && atAllVisible(view, rect) && n != 2 && view.willFocusWithin(rect, i)) {
	  if (view instanceof Container) 
	    if (n == 2 || tryToFocusWithin((Container)view, i, o, p, q, r)) {
	      container.setFocusedChild(view);
	      return (true);
	    }
                        
	    else {
	      container.setFocusedChild(view);
	      return (true);
	    }
	}
	if (i) {
	  view = container.getNextFocusableChild(view);
	  continue;
	}
	view = container.getPrevFocusableChild(view);
      }
      n++;
    }
    return (false);
  }


  public boolean pageUp()
  {
    int i;
    if (mOriginY == 0) {
      i = 0;
      Container container = this;
      while (container != null) {
	View view = MYgetFirstFocusableChild(container);
	if (view == null) {
	  container = null;
	  continue;
	}
	else if (this != container && !canChangeFocusTo(view, false)) 
	  break;
                   
	else if (!container.childIsOnScreen(view)) 
	  break;
                   
	else if (container.getFocusedChild() != view) {
	  container.setFocusedChild(view);
	  i = 1;
	}
	if (view instanceof Container) {
	  container = (Container)view;
	  continue;
	}
	container = null;
      }
      if (i != 0) 
	return (true);
              
      playLimitSound(false);
      return (false);
    }
    i = getHeight() - (mScrollDelta * 3>>2);
    offsetOriginAndSelect(i);
    getParent().invalidate();
    return (true);
  }


  public boolean pageDown()
  {
    int i;
    if (getHeight() > mScrollBottom || getHeight() - mScrollBottom == mOriginY) {
      i = 0;
      Container container = this;
      while (container != null) {
	View view = MYgetLastFocusableChild(container);
	if (view == null) {
	  container = null;
	  continue;
	}
	else if (this != container && !canChangeFocusTo(view, true)) 
	  break;
                   
	else if (!container.childIsOnScreen(view)) 
	  break;
                   
	else if (container.getFocusedChild() != view) {
	  container.setFocusedChild(view);
	  i = 1;
	}
	if (view instanceof Container) {
	  container = (Container)view;
	  continue;
	}
	container = null;
      }
      if (i != 0) 
	return (true);
              
      playLimitSound(true);
      return (false);
    }
    i = -getHeight() + (mScrollDelta * 3>>2);
    offsetOriginAndSelect(i);
    getParent().invalidate();
    return (true);
  }


  protected View getFirstChildBelowTop()
  {
    View view = getFirstChild();
    int i = -mOriginY;
    for (;view != null;) {
      if (view.getTop() >= i) 
	return (view);
              
      if (view.getTop() + view.getHeight() - mScrollDelta >= i) 
	return (view);
              
      view = view.getSibling();
    }
    return (null);
  }


  protected View getLastChildAboveBottom()
  {
    View view = getFirstChild();
    View view2 = null;
    int i = -mOriginY;
    int j = 0;
    i += getHeight();
    for (;view != null; view = view.getSibling()) {
      int l = view.getTop() + view.getHeight();
      if (l < j && l <= i || view.getTop() + mScrollDelta < i) 
	view2 = view;
    }
    return (view2);
  }


  protected View getFirstVisibleChild()
  {
    View view = getFirstChild();
    for (;view != null;) {
      if (atAllVisible(view)) 
	return (view);
              
      view = view.getSibling();
    }
    return (null);
  }


  private final boolean entirelyVisible(View view)
  {
    int i = -mOriginY;
    int j = i + getHeight();
    if (view.getTop() < i && view.getTop() + view.getHeight() <= j) 
      return (true);
         
    return (false);
  }


  private final boolean entirelyVisible(View view ,Rect rect)
  {
    if (view.getTop() < rect.top && view.getTop() + view.getHeight() <= rect.bottom) 
      return (true);
         
    return (false);
  }


  private final boolean atAllVisible(View view)
  {
    int i = -mOriginY;
    int j = i + getHeight();
    if (view.getTop() >= j && view.getTop() + view.getHeight() > i) 
      return (true);
         
    return (false);
  }


  private final boolean atAllVisible(View view ,Rect rect)
  {
    if (view.getTop() >= rect.bottom && view.getTop() + view.getHeight() > rect.top) 
      return (true);
         
    return (false);
  }


  protected boolean reverseFocusOrScroll(int i)
  {
    if (tryToReverseFocus()) 
      return (true);
         
    if (mOriginY == 0) {
      playLimitSound(false);
      return (false);
    }
    offsetOrigin(mScrollDelta * i);
    getParent().invalidate();
    return (true);
  }


  private final boolean tryToReverseFocus()
  {
    View view = getPrevFocusableChild();
    if (canChangeFocusTo(view, false)) {
      focusChangedTo(view, false);
      return (true);
    }
    for (;view != null;) {
      if (canChangeFocusTo(view, false)) {
	focusChangedTo(view, false);
	return (true);
      }
      view = getPrevFocusableChild(view);
    }
    return (false);
  }


  protected void playLimitSound(boolean i)
  {
    Container container = (Container)getParent();
    if (container == null || container.getFocusableChildInWindow(i) == null) 
      Meta.play(Meta.SCROLL_LIMIT);
         
    return;
  }


  public void setOrigin(int i ,int j)
  {
    if (mScrollBottom + j < getHeight()) 
      j = getHeight() - mScrollBottom;
         
    if (j > 0) 
      j = 0;
         
    super.setOrigin(i, j);
    return;
  }


  public boolean keepRectInView(View view ,Rect rect)
  {
    boolean changed = false;
    Rect rectInView = mWorkRect1;
    rectInView.left = rect.left;
    rectInView.top = rect.top;
    rectInView.right = rect.right;
    rectInView.bottom = rect.bottom;
    rectInView.translate(view.getLeft(), view.getTop());
    Rect viewableAreaRect = getBounds(mWorkRect2);
    viewableAreaRect.translate(-mOriginX, -mOriginY);
    if (rectInView.left < viewableAreaRect.left + mHorizontalScrollContext) {
      offsetOrigin((viewableAreaRect.left + mHorizontalScrollContext - rectInView.left), 0);
      changed = true;
    } 
    else if (rectInView.right > viewableAreaRect.right - mHorizontalScrollContext) {
      offsetOrigin((viewableAreaRect.right - mHorizontalScrollContext - rectInView.right), 0);
      changed = true;
    }
    if (rectInView.top < viewableAreaRect.top + mScrollContext) {
      offsetOrigin(0, viewableAreaRect.top + mScrollContext - rectInView.top);
      changed = true;
    }
    else if (rectInView.bottom > viewableAreaRect.bottom - mScrollContext) {
      offsetOrigin(0, (viewableAreaRect.bottom - mScrollContext - rectInView.bottom));
      changed = true;
    }
    if (changed)
      getParent().invalidate();
    return (changed);
  }

  public void setScrollDelta(int i)
  {
    mScrollDelta = i;
    return;
  }


  public void setScrollContext(int i)
  {
    mScrollContext = i;
    return;
  }

  public void setHorizontalScrollContext(int i)
  {
    mHorizontalScrollContext = i;
    return;
  }

  public void setHasScrollbar(boolean i)
  {
    mHasScrollbar = i;
    return;
  }


  public boolean getHasScrollbar()
  {
    return (mHasScrollbar);
  }


  public void paint(Pen pen)
  {
    super.paint(pen);
    if (mHasScrollbar) {
      pen.offset(getWidth() - mScrollbar.getWidth() - mOriginX, -mOriginY);
      updateScrollbar();
      mScrollbar.paint(pen);
    }
    return;
  }


  protected void updateScrollbar()
  {
    Rect rect = getBounds();
    mScrollbar.setSize(0, rect.getHeight() - 1);
    mScrollbar.setData(-mOriginY, rect.getHeight(), mScrollBottom);
    return;
  }


  public int getContentWidth()
  {
    int i = 0;
    if (mHasScrollbar) 
      i = mScrollbar.getWidth();
         
    return (getWidth() - i);
  }

  public MyScrollView()
  {
    this(0, 0, 0, 0);
    return;
  }


  public MyScrollView(Rect rect)
  {
    this(rect.left, rect.top, rect.right, rect.bottom);
    return;
  }


  public MyScrollView(int i ,int j ,int k ,int l)
  {
    super(i, j, k, l);
    setAcceptFocus(true);
    mWorkRect1 = new Rect();
    mWorkRect2 = new Rect();
    mWorkPoint = new Point();
    Font font = Font.findSystemFont();
    mScrollDelta = font.getAscent() + font.getDescent();
    mScrollbar = new Scrollbar(true);
    setHasScrollbar(true);
    return;
  }
  // --------------------------------
  protected View MYgetFirstFocusableChild(Container c)
  {
    View view = c.getFirstChild();
    while (view != null) {
      if (view.isFocusable()) 
	return (view);
      view = view.getSibling();
    }
    return (null);
  }


  protected View MYgetLastFocusableChild(Container c)
  {
    Object object = null;
    View view = c.getFirstChild();
    View view1 = view;
    while (view != null) {
      if (view.isFocusable()) 
	view1 = view;
      view = view.getSibling();
    }
    return (view1);
  }

}
