package com.hipme.flipsheet;

import danger.ui.TableCell;
import danger.ui.Pen;
import danger.ui.Rect;
import danger.ui.StaticText;
import danger.ui.Font;
import danger.util.DEBUG;

/**
/** This class should not have to exist, 
 * but there are bugs in danger.ui.TableCell in hiptop OS 1.1
 * This class reproduces enough of danger.ui.TableCell's functionality
 * that com.hipme.SheetTableCell can specialize it, and still be
 * forward compatible when Danger fixes TableCell.
 * 
 * This class clips the contents instead of allowing them to draw outside the rect.
 * As an optimization, the clipping is enabled only if it looks like it will be needed according
 * to font.getWidth(), but that may be a false optimization and it may be cheaper to
 * clip always and not do the stringwidth calculation.
 */
public class MyTableCell extends TableCell {
  protected SheetTable sheetTable;
  // later we'd like to allow this to be set to false
  // if the followign cell is empty, but then
  // we also need to have the following cell not
  // draw its borders.
  final static boolean clip = true; 

  MyTableCell() {
    super();
  }
  MyTableCell(SheetTable table) {
    super(table);
    sheetTable = table;
  }

  public void paint(Pen pen ,Rect rect) {
    Font font;
    int j;
    int i;
    int width = getWidth();
    int right = rect.left + width;
    int bottom = rect.top + getHeight();
    if (getBackgroundColor() != -1) {
      pen.setColor(getBackgroundColor());
      pen.fillRect(rect.left , rect.top , right , bottom);
    }
    pen.setColor(getForegroundColor());
    // crock
    font = sheetTable.valueFont;
         
    i = font.getAscent();

    String mStringValue = getElementAsString();
    if (mStringValue != null) {
      int mAlign = getAlignment();
      int k = font.getWidth(mStringValue);
      boolean clipNeeded = (k > width);
      if (clip && clipNeeded && mAlign == StaticText.HALIGN_RIGHT) {
	// If align left, assume numeric and instead of truncating,
	// show "###".  We don't adjust width variable becasue
	// it's not used below here.
	mStringValue = "###";
	k = font.getWidth(mStringValue);
	clipNeeded = (k > width);
      }
      // Now go on.
      if (mAlign == StaticText.HALIGN_LEFT) {
	j = rect.left + 1;
      } else if (mAlign == StaticText.HALIGN_RIGHT) {
	j = rect.left + width - k - 1;
      } else {
	// StaticText.HALIGN_CENTER
	j = rect.left + (width - k) / 2;
      }
      pen.setFont(font);
      if (clip && clipNeeded) {
	pen.push();
	pen.setClip(rect.left, rect.top, right, bottom);
      }
      try {
	pen.drawText(j , rect.top + i , mStringValue);
      } finally {
	if (clip && clipNeeded)
	  pen.pop();
      }
    }
  }
}
