package com.hipme.flipsheet;

import danger.app.*;
import danger.app.Application;
import danger.audio.Meta;
import danger.internal.Date;
import danger.ui.*;
import danger.ui.Pen;
import danger.system.*;
import danger.util.DEBUG;
import java.util.Vector;

final class SheetScrollView extends MyScrollView implements Events, Constants {
  SheetTable sheetTable;
  int w, h;
  int x0 = 0;
  int y0 = 0;

  boolean tiled = true;
  boolean isTiled() { return tiled; }  
  
  SheetScrollView(int w, int h) {
    super();
    resize(w, h);
  }

  void resize (int w, int h) {
    this.w = w;
    this.h = h;
    x0 = 0; 
    y0 = 0;
    setTiled(isTiled());
  }

  void setTiled(boolean tiled) {
    int myW = w;
    int myH = h;
    int myX = 0;
    int myY = 0;
    if (tiled) {
      myY += CellContentsTextField.FIELD_HEIGHT;
      myH -= CellContentsTextField.FIELD_HEIGHT;
    }
    setPosition(myX, myY);
    setSize(myW, myH);
    setOrigin(x0, y0);
    setActualSize(myW, myH);
    this.tiled = tiled;
  }

  void setSheetTable(SheetTable sheetTable) {
    this.sheetTable = sheetTable;
    removeAllChildren();
    addChild(sheetTable);
    Font font = Font.findSystemFont();
    setScrollDelta(2 * 
		   (font.getAscent() + font.getDescent() + 
		    sheetTable.getBorderWidth() + sheetTable.getCellSpacing() + 
		    sheetTable.getCellPadding()));
    resize(w, h);		// get w,h from table later.
  }

  void sendUpdateCell() {
    new Event(getParent(), kUpdateCell).sendToTarget();
  }

  void sendEditCellNoSelect() {
    Event e = new Event(getParent(), kEditCell);
    e.what = WHAT_NO_MOTION;
    e.sendToTarget();
  }

  void sendExitApp() {
    new Event(Application.getCurrentApp(), kReturnToLauncher).sendToTarget();
  }

  public boolean eventWidgetUp(int i, Event event) {
    boolean consumed = true;
    switch(i) {
    case Event.DEVICE_WHEEL_BUTTON:
      sendEditCellNoSelect();
      break;
    case Event.DEVICE_WHEEL:
    case Event.DEVICE_MULTIPLE_WHEEL:
      if (sheetTable.isWheelAxisHorizontal())
	wheelLeft(event.getWheelCount());
      else
	wheelUp(event.getWheelCount());
      break;
    default:
      consumed = super.eventWidgetUp(i, event);
    }
    return consumed;
  }

  /**
   * @returns true if successful, false if not.
   */
  boolean doScrollRight() {
    DEBUG.p("doScrollRight");
    ((flipsheet)Application.getCurrentApp()).exactTableWidth = false;
    // Rect r = sheetTable.getCellBounds(new Rect());
    x0 -= 10;
    DEBUG.p("setOrigin("+x0+",y0)");
    invalidate();
    setOrigin(x0, y0);
    invalidate();
    return true;
  }

  boolean doScrollLeft() {
    // Rect r = sheetTable.getCellBounds(new Rect());
    ((flipsheet)Application.getCurrentApp()).exactTableWidth = false;
    if (x0 == 0) 
      return false;
    x0 += 10;
    if (x0 > 0) 
      x0 = 0;
    DEBUG.p("setOrigin("+x0+",y0)");
    invalidate();
    setOrigin(x0, y0);
    invalidate();
    return true;
  }


  public boolean eventWidgetDown(int i, Event event) {
    boolean consumed = true;
    switch(i) {
    case Event.DEVICE_WHEEL:
    case Event.DEVICE_MULTIPLE_WHEEL:
	if (sheetTable.isWheelAxisHorizontal()) {
	  wheelRight(event.getWheelCount());
	} else {
	  wheelDown(event.getWheelCount());
	}
      break;
    case Event.DEVICE_ARROW_RIGHT:
      if (event.modifierIsActive(Event.EVENT_MODIFIER_MENU_BUTTON)) { 
	doScrollRight();
      } else {
	sheetTable.nextCol();
	sendUpdateCell();
      }
      if (true)
	keepCellInView();
      break;
    case Event.DEVICE_ARROW_LEFT:
      if (event.modifierIsActive(Event.EVENT_MODIFIER_MENU_BUTTON)) {
	doScrollLeft();
      } else {
	sheetTable.previousCol();
	sendUpdateCell();
      }
      if (true)
	keepCellInView();
      break;
    case Event.DEVICE_ARROW_DOWN:
      arrowDown();
      break;
    case Event.DEVICE_ARROW_UP:
      arrowUp();
      break;
    case Event.DEVICE_BUTTON_BACK:
      // X exits app
      sendExitApp();
      break;
    default:
      consumed = super.eventWidgetDown(i, event);
    }
    return consumed;
  }

  private void arrowUp() {
    sheetTable.previousRow();
    sendUpdateCell();
    keepCellInView();
  }

  private void arrowDown() {
    sheetTable.nextRow();
    sendUpdateCell();
    keepCellInView();
  }


  private void wheelUp(int rows) {
    for (int i = 0; i < rows; i++)
      sheetTable.previousRow();
    sendUpdateCell();
    keepCellInView();
  }

  private void wheelDown(int rows) {
    for (int i = 0; i < rows; i++)
      sheetTable.nextRow();
    sendUpdateCell();
    keepCellInView();
  }

  private void wheelLeft(int cells) {
    boolean rowChanged = false;
    for (int i = 0; i < cells; i++) {
      if (sheetTable.backCell()) 
	rowChanged = true;
    }
    sendUpdateCell();
    if (true || rowChanged)
      keepCellInView();
  }

  private void wheelRight(int cells) {
    boolean rowChanged = false;
    for (int i = 0; i < cells; i++) {
      if (sheetTable.forwardCell())
	rowChanged = true;
    }
    sendUpdateCell();
    if (true || rowChanged)
      keepCellInView();
  }

  private Rect rectToKeepInView = new Rect();  // Is this a good idea?

  void keepCellInView() {
    keepRectInView(sheetTable, sheetTable.getInterestingRect(rectToKeepInView));
  }
}
