// Copyright 2003, 2004 Leigh L. Klotz, Jr.
// All rights reserved.

package com.hipme.flipsheet;

import danger.app.*;
import danger.app.IPCMessage;
import danger.app.Registrar;
import danger.app.SettingsDB;
import danger.audio.Meta;
import danger.internal.Date;
import danger.net.HiptopConnection;
import danger.net.HiptopConnection;
import danger.net.HiptopTransaction;
import danger.net.URL;
import danger.net.URLException;
import danger.system.*;
import danger.ui.*;
import danger.ui.AlertWindow;
import danger.ui.CheckBox;
import danger.ui.Color;
import danger.ui.DialogWindow;
import danger.ui.NumberEditor;
import danger.ui.EditText;
import danger.ui.StaticText;
import danger.ui.TextField;
import danger.ui.TextInputAlertWindow;
import danger.util.DEBUG;
import danger.util.format.StringFormat;
import java.util.Vector;

/**
 * Implements the main application class.
 *
 */
public final class flipsheet extends Application implements Resources, Events
{
  final static String CSV_DATASTORE_NAME = "CSV";
  final static int MAX_COLOR_SHEETS = 40;
  final static int MAX_GRAY_SHEETS = 40;
  final static int MAX_SHEETS = (Screen.isColor() ? MAX_COLOR_SHEETS : MAX_GRAY_SHEETS);
  final static boolean attachmentsSupported = true;

  private static SheetWindow sheetWindow;
  int sheetNumber = 0;
  int sheetUID = -1;

  final static String SETTINGS_DB="flipsheet";
  SettingsDB settings;
  final static String SHOW_DESCENDERS = "showDescenders";
  boolean showDescenders = false;
  final static String USE_OUTLINE = "useOutline";
  boolean useOutline = !Screen.isColor();
  final static String ACCEPT_FLIPS = "acceptFlips";
  boolean acceptFlips = true;

  boolean exactTableWidth = false; // FIXME

  /**
   * Creates the main application class.
   *
   */
  public flipsheet() {
    super();

    // init data store
    DataStore ds = getDataStore();
    if (ds == null) {
      DEBUG.p("flipsheet: no datastore found");
      ds = DataStore.createDataStore(CSV_DATASTORE_NAME, true);
      DEBUG.p("flipsheet: ds = " +ds);
      // Probably want to ask the user before doing this...
      ds.addRecord(getString(kDefaultInitialFlipsheetCSV1).getBytes());
      ds.addRecord(getString(kDefaultInitialFlipsheetCSV2).getBytes());
      DEBUG.p("count = " + ds.getRecordCount());
      new Event(this, kWelcomeEvent).sendToTarget();
    }
    ds.setAutoSyncNotifyee(this);
    DataStore.registerListenerForAutoSyncDone(this);

    // init settings
    boolean oldSettings = SettingsDB.findDB(SETTINGS_DB);
    settings = new SettingsDB(SETTINGS_DB, true);
    if (oldSettings) {
      restoreSettings();
    } else {
      initSettings();
    }
    // Registrar.registerURLSchemeHandler("flipsheet:");
  }


  private void initSettings() {
    settings.setIntValue(SHOW_DESCENDERS, showDescenders ? 1 : 0);
    settings.setIntValue(USE_OUTLINE, useOutline ? 1 : 0);
    settings.setIntValue(ACCEPT_FLIPS, acceptFlips ? 1 : 0);
  }

  private void restoreSettings() {
    try { showDescenders = (settings.getIntValue(SHOW_DESCENDERS) != 0); } 
    catch (SettingsDBException settingsdbexception) { }
    try { useOutline = (settings.getIntValue(USE_OUTLINE) != 0); } 
    catch (SettingsDBException settingsdbexception) {
      useOutline = !Screen.isColor(); 
    }
    try { acceptFlips = (settings.getIntValue(ACCEPT_FLIPS) != 0); } 
    catch (SettingsDBException settingsdbexception) { }
  }

  /**
   * Handles the Suspend event.
   */
  public void suspend() {
    saveSheet();
    removeEmptySheets();
    super.suspend();
  }


  public boolean receiveEvent(Event e) {
    boolean consumed = true;
    switch (e.type) {
    case kReturnToLauncher:
    case kLockedBundleExit:
      returnToLauncher();
      break;
    case kSetOnLoadEvent:
      doSetOnLoad(e);
      break;
    case kCSVEditDoneEvent:
      doCSVEditDone(e);
      break;
    case kSheetSettingsEvent:
      doSheetSettings(e);
      break;      
    case kCellSettingsEvent:
      doCellSettings(e);
      break;
    case kShowColumnSettingsEvent:
      showColumnSettings();
      break;
    case kShowLockedBundleAlert:
      showLockedBundleAlert();
      break;
    case kShowExpiredBundleAlert:
      showExpiredBundleAlert();
      break;
    case kDisplaySettingsDoneEvent:
      doDisplaySettingsDoneEvent(e);
      break;
    case kAboutEvent:
      showAbout();
      break;
    case kWelcomeEvent:
      showWelcome();
      break;
    case kHelpEvent:
      showHelp();
      break;
    case kImportFlipSheetEvent:
      doImportFlipSheet(e);
      break;
    case kDisplaySettingsEvent:
      showDisplaySettings();
      break;
    case kShowCellSettingsEvent:
      showCellSettingsDialog();
      break;
    case kConfirmDiscardSheetEvent:
      showAlert(kConfirmDiscardSheetAlert, this);
      break;
    case kDiscardSheetEvent:
      doDiscardSheet(e);
      break;
    case kInsertRowsEvent:
      doInsertRows(e);
      break;
    case kShowInsertColumnsDialogEvent:
      showInsertColumnsDialog();
      break;
    case kInsertColumnsEvent:
      doInsertColumns(e);
      break;
    case kConfirmDiscardRowsEvent:
      showDialog(kConfirmDiscardRowsDialog, this);
      break;
    case kDiscardRowsEvent:
      doDiscardRows(e);
      break;
    case kConfirmDiscardColumnsEvent:
      showDialog(kConfirmDiscardColumnsDialog, this);
      break;
    case kDiscardColumnsEvent:
      doDiscardColumns(e);
      break;
    case kShowFlipOutFlipSheetEvent:
      showTextInputAlert(kFlipOutFlipSheetTIA, this);
      break;
    case kEditOnLoad:
      showOnLoadEditor();
      break;
    case kShowSheetSettingsEvent:
      showSheetSettingsDialog();
      break;      
    case kFlipOutFlipSheetEvent:
      doFlipOutFlipSheet(e);
      break;
    case kFlipInFlipSheetEvent:
      doFlipInFlipSheet(e);
      break;
    case kMailFlipSheetEvent:
      doMailSheet(e);
      break;
    case kNewFlipSheet:
      doNewFlipSheet(e);
      break;
    case kSetFlipSheet:
      doSetSheetNumber(e);
      break;
    default:
      consumed = super.receiveEvent(e);
    }
    return consumed;
  }


  private void showLockedBundleAlert() {
    AlertWindow a = getAlert(kLockedBundleAlert, this);
    a.setMessage(StringFormat.fromRsrc(kLockedBundleAlertText, getString(kLockedBundleUser)));
    a.show();
  }

  private void showEvaluationBundleAlert() {
    AlertWindow a = getAlert(kEvaluationBundleAlert, this);
    a.setMessage(StringFormat.fromRsrc(kEvaluationBundleAlertText, getString(kExpiredBundleDate)));
    a.setAutoDismiss(1000, null);
    a.show();
  }

  private void showExpiredBundleAlert() {
    AlertWindow a = getAlert(kExpiredBundleAlert, this);
    a.setMessage(StringFormat.fromRsrc(kEvaluationBundleAlertText, getString(kExpiredBundleDate)));
    a.show();
  }

  /**
   * Handles the Resume event.
   *
   */
  public void resume() {
    System.err.println("flipsheet: Version = " + getString(ID_VERSION));
    DataStore ds = getDataStore();
    // See if records got reordered.
    boolean mismatch = (ds != null) && (ds.getRecordUID(sheetNumber) != sheetUID);
    if ((sheetWindow != null) && mismatch)
      sheetWindow.invalidateSheetNames();
    // Create/get the main window...
    if (sheetWindow == null || mismatch) {
      // Get the data store and read the CSV
      sheetWindow = new SheetWindow(getCSV());
      sheetWindow.setTitle(getString(kStrSheetWindowTitle));
      if (ds != null) 
	sheetUID = ds.getRecordUID(sheetNumber);
    }
    // Show the main window...
    sheetWindow.update();
    sheetWindow.setTitle(getSheetTable().getSheetName());   // why do I have to do this?
    sheetWindow.show();

    // check for locked bundle
    String user = (getString(kLockedBundleUser));
    if (user.length()>0 && !HiptopConnection.getUserName().startsWith(user)) {
      new Event(this, kShowLockedBundleAlert).sendToTarget();
    }
    // check for expired
    {
      String expirationDate = getString(kExpiredBundleDate);
      if (expirationDate.length() > 1) { // has to be >1 because drc fails on empty string sometiems
	if (new Date().compareTo(new Date(2004, 12, 31)) >= 0) { // crock for missing date parser
	  new Event(this, kShowExpiredBundleAlert).sendToTarget();
	} else {
	  showEvaluationBundleAlert();	// just blink the alert briefly
	}
      }
    }
  }
	
  DataStore getDataStore() {
    return DataStore.findDataStore(CSV_DATASTORE_NAME);
  }


  /**
   * Commit edits, and save sheet as CSV, even if empty.
   */
  void saveSheet() {
    putCSV(sheetWindow.getCSV());
  }

  /**
   * On suspend, flush any empty sheets.  We don't do this while the app is running
   * to avoid re-ordering the display when someone creates a new sheet and navigates
   * off it.
   * Remove any empty sheets (except 0 and current sheet).
   * Updates sheetNumber without redisplay so call only on suspend()
   * or when you know the sheet is not displayed.
   */
  private void removeEmptySheets() {
    DataStore ds = getDataStore();
    if (ds != null) {
      for (int i = 1; i < getSheetCount(); i++) {
	if (i != sheetNumber) {
	  // there should be a faster way of tell what was empty in this session.
	  if (ds.getRecordData(i).length == 0) {
	    DEBUG.p("flipsheet: Dropping empty record " + i);
	    ds.removeRecord(i);
	    // move sheet number down behind its back.
	    if (i < sheetNumber) sheetNumber--;
	  }
	}
      }
    }
  }

  void putCSV(String csv) {
    DataStore ds = getDataStore();
    DEBUG.p("flipsheet: putting CSV to " + ds);
    if (ds == null) throw new RuntimeException("Cannot happen -- no DataStore " + CSV_DATASTORE_NAME);
    if (ds.getRecordCount() <= sheetNumber) {
      int rn = ds.addRecord(csv.getBytes());
    } else {
      ds.setRecordData(sheetNumber, csv.getBytes());
    }
    ds.discardOnReboot(false);	// should have done this when making it.
    DEBUG.p("flipsheet: Record " + sheetNumber + " set: total bytes = " + ds.getBytesUsed());
  }

  /**
   * Returns CSV string or null if none.
   */

  String getCSV() {
    return getCSV(sheetNumber);
  }

  String getCSV(int sn) {
    String result = null;
    DataStore ds = getDataStore();
    byte[] rec = ds.getRecordData(sn);
    // We get no records if we ask for data before the connection is up. 
    // How do we prevent this?
    if (rec != null) {
      result = new String(rec);
    } else {
      DEBUG.p("flipsheet: Got empty CSV result for " + sn);
    }
    return result;
  }

  /**
   * Returns true if succesful and selected,
   * false if full.
   */
  boolean makeNewSheet() {
    saveSheet();
    if (getSheetCount() >= MAX_SHEETS) return false;
    setSheetNumber(getSheetCount());
    return true;
  }

  private void doSetSheetNumber(Event e) {
    setSheetNumber(e.data);
  }

  /**
   * Makes sure current sheet is saved, even if it's never been
   * written to the store before, and switches to the specified
   * sheet, making it if necessar.
   */
  void setSheetNumber(int i) {
    if (i == sheetNumber) return;
    if (i < 0 || i >= MAX_SHEETS) {
      DEBUG.p("flipsheet: sheet number out of range: " + i);
      Meta.play(Meta.BEEP_ACTION_FAIL);
      return;
    }
    saveSheet();
    setSheetNumberTail(i);
  }

  /** Switches to the specified sheet, making it if necessary.
   */
  private void setSheetNumberTail(int i) {
    sheetNumber = i;
    sheetWindow.update(getCSV());
    if (sheetNumber >= getSheetCount()) {
      saveSheet();
    }
    DataStore ds = getDataStore();
    if (ds != null) 
      sheetUID = ds.getRecordUID(sheetNumber);
    sheetWindow.setTitle(getSheetTable().getSheetName());
    Meta.play(Meta.BEEP_ACTION_SUCCESS);    
  }

  /**
   * Call to delete current sheet.
   */
  void doDiscardSheet(Event e) {
    int i = sheetNumber;
    DataStore ds = getDataStore();
    if (ds != null) {
      DEBUG.p("flipsheet: deleting sheet record " + i);
      ds.removeRecord(i);
      Meta.play(Meta.DELETE_ITEM);
      setSheetNumberTail(i == 0 ? 0 : (i-1));
    }
  }

  /**
   * Call to insert rows at current row,
   * leaving current row as first new blank row.
   */
  void doInsertRows(Event e) {
    sheetWindow.addDiscardRowsColumns(1, 0, false, false);
  }

  /**
   * Call to delete rows at current row,
   * including current row.  Should be confirmed first!
   */
  void doDiscardRows(Event e) {
    int nRows = ((NumberEditor)((DialogWindow)e.argument).getDescendantWithID(kDiscardRowsRows)).getValue();
    sheetWindow.addDiscardRowsColumns(-nRows, 0, false, false);
    Meta.play(Meta.DELETE_ITEM);
  }

  /**
   * Call to delete columns at current column,
   * including current column.  Should be confirmed first!
   */
  void doDiscardColumns(Event e) {
    int nColumns = ((NumberEditor)((DialogWindow)e.argument).getDescendantWithID(kDiscardColumnsColumns)).getValue();
    sheetWindow.addDiscardRowsColumns(0, -nColumns, false, false);
    Meta.play(Meta.DELETE_ITEM);
  }

  void showInsertColumnsDialog() {
    DialogWindow dw = getDialog(kInsertColumnsDialog, this);
    dw.show();
  }

  /**
   * Call to insert columns at current column,
   * splitting width with current column and inserting
   * new column to the right.
   */
  void doInsertColumns(Event e) {
    DialogWindow dw = ((DialogWindow)e.argument);
    int how = ((PopupMenu)(dw.getDescendantWithID(kInsertColumnsDialogHowMenuPopupWidget))).getValueAsItemID();
    int where = ((PopupMenu)dw.getDescendantWithID(kInsertColumnsDialogWhereMenuPopupWidget)).getValueAsItemID();
    boolean insertOnLeft = (where == kInsertColumnsDialogWhereOnLeft);
    boolean makeTableWider = (how == kInsertColumnsDialogHowMakeTableWider);
    sheetWindow.addDiscardRowsColumns(0, 1, insertOnLeft, makeTableWider);
  }

  private void showAbout() {
    AlertWindow a = getAlert(kAboutAlert, this);
    String user = getString(kLockedBundleUser);
    String v = (user.length()==0) ? getString(ID_VERSION) : getString(ID_VERSION) + "-"+ user;
    a.setMessage(StringFormat.fromRsrc(kAboutAlertText, v));
    a.show();
  }

  private void showWelcome() {
    AlertWindow a = getAlert(kWelcomeAlert, this);
    a.show();
  }

  void showHelp() {
    try {
      URL.gotoURL(getString(kHelpURL) + getString(ID_VERSION));
    }
    catch (URLException e) {
      SheetWindow.errorprint(e);
    }
  }

  //
  // Display Settings
  //
  private void showDisplaySettings() {
    DialogWindow settingsDialog = getDialog(kDisplaySettingsDialog, this);
    boolean compact = !showDescenders;
    ((CheckBox)(settingsDialog.getDescendantWithID(kDisplaySettingsCompactCheckBox))).setValue(compact ? 1 : 0);
    ((CheckBox)(settingsDialog.getDescendantWithID(kDisplaySettingsUseOutlineCheckBox))).setValue(useOutline ? 1 : 0);
    ((CheckBox)(settingsDialog.getDescendantWithID(kDisplaySettingsAcceptFlipsCheckBox))).setValue(acceptFlips ? 1 : 0);
    settingsDialog.show();
  }

  private void doDisplaySettingsDoneEvent(Event e) {
    DialogWindow settingsDialog = (DialogWindow)(e.argument);
    SheetTable sheetTable = getSheetTable();
    {
      int compact = (((CheckBox)(settingsDialog.getDescendantWithID(kDisplaySettingsCompactCheckBox))).getValue());
      showDescenders = (compact == 0);
      settings.setIntValue(SHOW_DESCENDERS, showDescenders ? 1 : 0);
      sheetTable.setShowDescenders(showDescenders);
    }
    {
      int val = (((CheckBox)(settingsDialog.getDescendantWithID(kDisplaySettingsUseOutlineCheckBox))).getValue());
      useOutline = (val != 0);
      settings.setIntValue(USE_OUTLINE, useOutline ? 1 : 0);
      sheetTable.setUseOutline(useOutline);
    }
    {
      int val = (((CheckBox)(settingsDialog.getDescendantWithID(kDisplaySettingsAcceptFlipsCheckBox))).getValue());
      acceptFlips = (val != 0);
      settings.setIntValue(ACCEPT_FLIPS, acceptFlips ? 1 : 0);
    }
  }

  //
  // Evaluate On Load
  //
  private void showOnLoadEditor() {
    DialogWindow onLoadEditor = getDialog(kEditOnLoadDialog, this);
    EditText t = ((EditText)(onLoadEditor.getDescendantWithID(kEditOnLoadTextID)));
    t.setText(getSheetTable().getOnLoadText());
    onLoadEditor.show();
  }

  private void doSetOnLoad(Event e) {
    DialogWindow onLoadEditor = (DialogWindow)(e.argument);
    EditText t = ((EditText)(onLoadEditor.getDescendantWithID(kEditOnLoadTextID)));
    if (t == null) return;
    getSheetTable().setOnLoadText(t.toString());
  }

  //
  // Import
  //
  private void showCSVEditor() {
    String csv = sheetWindow.getCSV();
    if (csv == null) csv = "";
    DialogWindow editCSVEditor = getDialog(kEditCSVDialog, this);
    EditText t = ((EditText)(editCSVEditor.getDescendantWithID(kEditCSVTextID)));
    t.setText(csv);
    t.selectAll();
    editCSVEditor.show();
  }

  private void doCSVEditDone(Event e) {
    DialogWindow editCSVEditor = (DialogWindow)(e.argument);
    EditText t = ((EditText)(editCSVEditor.getDescendantWithID(kEditCSVTextID)));
    if (t == null) return;
    sheetWindow.update(t.getStringCRLF());
    saveSheet();
  }

  //
  // Sheet properties
  //
  private void showSheetSettingsDialog() {
    SheetTable sheetTable = getSheetTable();
    DialogWindow d = getDialog(kSheetSettingsDialog, this);
    {
      TextField t = ((TextField)(d.getDescendantWithID(kSheetSettingsDialogName)));
      t.setText(sheetTable.getSheetName());
      t.selectAll();
    }
    {
      NumberEditor t = ((NumberEditor)(d.getDescendantWithID(kSheetSettingsDialogPrecision)));
      t.setValue(sheetTable.getPrecision());
    }
    {
      NumberEditor n = ((NumberEditor)(d.getDescendantWithID(kSheetSettingsDialogSheetNumber)));
      n.setValue(getSheetNumber());
      n.max(getSheetCount()-1);
    }
    d.show();
  }

  private void doSheetSettings(Event e) {
    DialogWindow dialog = (DialogWindow)(e.argument);
    SheetTable sheetTable = getSheetTable();
    String name = ((TextField)(dialog.getDescendantWithID(kSheetSettingsDialogName))).toString();
    int precision = ((NumberEditor)(dialog.getDescendantWithID(kSheetSettingsDialogPrecision))).getValue();
    int sn = ((NumberEditor)(dialog.getDescendantWithID(kSheetSettingsDialogSheetNumber))).getValue();
    String oldDefaultName = StringFormat.fromRsrc(kStrSheetFormat, ("" + sheetNumber));
    // DEBUG.p("oldDefaultName = " + oldDefaultName);
    if (name.length()==0 || 
	((sn != sheetNumber) && sheetTable.getSheetName().equals(oldDefaultName))) {
      name = StringFormat.fromRsrc(kStrSheetFormat, ("" + sn));
    }
    // DEBUG.p("Setting name to " + name);
    sheetWindow.setSheetName(name);
    sheetTable.setPrecision(precision);
    if (sheetNumber != sn) {
      // DEBUG.p("before sheetNumber="+sheetNumber+" sheetUID="+sheetUID);
      // DEBUG.p("Swapping sheet " + sheetNumber + " swapped with " + sn);
      getDataStore().swapRecords(sheetNumber, sn);
      // update sheet number behind our own back.
      sheetNumber = sn;
      sheetUID = getDataStore().getRecordUID(sheetNumber);
      sheetWindow.invalidateSheetNames();
      // DEBUG.p("after sheetNumber="+sheetNumber+" sheetUID="+sheetUID);
    }
    // a bit harsh but ok for now
    sheetTable.calculate();
  }

  //
  // Cell Settings
  //
  private void showCellSettingsDialog() {
    DialogWindow dialog = getDialog(kCellSettingsDialog, this);
    SheetTable sheetTable = getSheetTable();
    {
      CheckBox t = ((CheckBox)(dialog.getDescendantWithID(kCellSettingsDialogLocked)));
      t.setValue(sheetTable.isCellLocked() ? 1 : 0);
    }
    {
      TextField t = ((TextField)(dialog.getDescendantWithID(kCellSettingsDialogFG)));
      t.setText(colorToString(((SheetTableCell)(sheetTable.getCell())).foregroundColor));
    }
    {
      TextField t = ((TextField)(dialog.getDescendantWithID(kCellSettingsDialogBG)));
      t.setText(colorToString(((SheetTableCell)(sheetTable.getCell())).backgroundColor));
    }
    dialog.show();
  }

  /**
   * When confirming cell properties, if any have changed,
   * add them to the Evaluate On Load list and eval them then.
   * Some day addOnLoadText will eliminate dups and flip/flops.
   */
  private void doCellSettings(Event e) {
    DialogWindow dialog = (DialogWindow)(e.argument);
    SheetTable sheetTable = getSheetTable();
    String cellName = sheetTable.getCellName();
    {
      CheckBox t = ((CheckBox)(dialog.getDescendantWithID(kCellSettingsDialogLocked)));
      boolean lock = (t.getValue() != 0);
      if (sheetTable.isCellLocked() != lock) {
	sheetTable.addOnLoadText("lock("+t.getValue()+", " + cellName +")");
      }
    }
    {
      TextField t = ((TextField)(dialog.getDescendantWithID(kCellSettingsDialogFG)));
      String fg = t.toString();
      if (!Color.colorsEqual(((SheetTableCell)(sheetTable.getCell())).foregroundColor, Color.fromString(fg))) {
	// hack: setcolor understands #<symbol green> in addition to String "green",
	// which means we don't have to figure out if this is a number or a #hex number or a color name.
	sheetTable.addOnLoadText("setcolor("+fg+", " + cellName +")");
      }
    }
    {
      TextField t = ((TextField)(dialog.getDescendantWithID(kCellSettingsDialogBG)));
      String bg = t.toString();
      if (!Color.colorsEqual(((SheetTableCell)(sheetTable.getCell())).backgroundColor, Color.fromString(bg))) {
	sheetTable.addOnLoadText("setbg("+bg+", " + cellName+")");
      }
    }
  }

  //
  // Column Settings
  //
  private void showColumnSettings() {
    getDialog(kColumnSettingsDialog, sheetWindow).show();
  }

  void showCircularDependency(String message) {
    AlertWindow a = getAlert(kCircularDependencyAlert, this);
    a.setMessage(message);
    a.show();
    Meta.play(Meta.BEEP_ACTION_FAIL);
  }

  String colorToString(int color) {
    switch(color & 0xffffff) {
    case Color.BLACK:
	return "black";
    case Color.WHITE:
	return "white";
    case Color.RED:
	return "red";
    case Color.GREEN:
	return "green";
    case Color.BLUE:
	return "blue";
    case Color.CYAN:
	return "cyan";
    case Color.MAGENTA:
	return "magenta";
    case Color.ORANGE:
	return "orange";
    case Color.PURPLE:
	return "purple";
    case Color.VIOLET:
	return "violet";
    case Color.YELLOW:
	return "yellow";
    case Color.PINK:
	return "pink";
    case Color.GRAY1:
	return "gray1";
    case Color.GRAY2:
	return "gray2";
    case Color.GRAY3:
	return "gray3";
    case Color.GRAY4:
	return "gray4";
    case Color.GRAY5:
	return "gray5";
    case Color.GRAY6:
	return "gray6";
    case Color.GRAY7:
	return "gray7";
    case Color.GRAY8:
	return "gray8";
    case Color.GRAY9:
	return "gray9";
    case Color.GRAY10:
	return "gray10";
    case Color.GRAY11:
	return "gray11";
    case Color.GRAY12:
	return "gray12";
    case Color.GRAY13:
	return "gray13";
    case Color.GRAY14:
	return "gray14";
    default:
      return '#' + Color.toString(color);
    }
  }


  //
  // Flipping
  //
  private void doFlipOutFlipSheet(Event e) {
    flipOutFlipSheet(getCSV(), 
		     TextInputAlertWindow.getTextFieldValue((IPCMessage)(e.argument), 
							    (kFlipOutFlipSheetTIAUserName)));
  }

  /**
   * doFlipOutFlipSheet is called to flip a sheet using HiptopConnection.
   */
  private void flipOutFlipSheet(String csv, String toUserName) {
    // Rather than storing the seqNo in a data store, we just guarantee uniqueness
    // to within about 50 days by using bottom 32 bits of the time.
    // 2^32/84600= about 50.
    int seqNo = (int)(System.currentTimeMillis() & 0xffffffff);
    
    DEBUG.p("flipping flipsheet to " + toUserName+" seqNo="+seqNo);
    HiptopConnection.send(toUserName, csv.getBytes(), seqNo);
  }


  /**
   * doFlipInFlipSheet is called when the user presses accept
   * in the DialogWindow displaying a flipped in flipsheet.
   */
  private void doFlipInFlipSheet(Event e) {
    DialogWindow dialog = (DialogWindow)(e.argument);
    EditText et = ((EditText)(dialog.getDescendantWithID(kFlipInFlipSheetDialogCSV)));
    String csv = et.toString();
    if (makeNewSheet()) 
      sheetWindow.update(csv);
    else
      mailSheet(getString(kFlipInSheetOverloadBody) + csv, getString(kFlipInSheetOverloadSubject));
  }

  public final void networkEvent(Object item) {
    if (item instanceof HiptopTransaction) {
      HiptopTransaction ht = ((HiptopTransaction)item);
      String u = ht.getFromUsername();
      if (acceptFlips) {
	showFlipInFlipSheet(u, ht.getString());
      } else {
	DEBUG.p("flipsheet: not accepting flip from " + u);
      }
    } else {
      super.networkEvent(item);
    }
  }

  public final void showFlipInFlipSheet(String fromUsername, String csv) {
    DEBUG.p("flipsheet: received flipsheet from " + fromUsername);
    Meta.play(Meta.BEEP_ACTION_SUCCESS);
    DialogWindow dialog = getDialog(kFlipInFlipSheetDialog, sheetWindow);
    StaticTextBox t = ((StaticTextBox)(dialog.getDescendantWithID(kFlipInFlipSheetDialogFrom)));
    t.setText(StringFormat.fromRsrc(kStrFlipFrom, fromUsername));
    EditText e = ((EditText)(dialog.getDescendantWithID(kFlipInFlipSheetDialogCSV)));
    e.setText(csv);
    e.setInsertionPoint(0);
    dialog.show();
  }

  //
  // Mail Flipsheet
  //
  private void doImportFlipSheet(Event e) {
    showCSVEditor();
    sheetWindow.invalidateSheetNames();
  }

  private void doMailSheet(Event e) {
    mailSheet(sheetWindow.getCSV(), 
	      StringFormat.fromRsrc(kStrMailSheetFormat, sheetWindow.getTitle()));
  }

  private void mailSheet(String csv, String subject) {
    IPCMessage ipc = new IPCMessage();
    ipc.addItem("action" , "send");
    if (subject != null) 
      ipc.addItem("subject" , subject);
    if (attachmentsSupported) {
      String [] names = { subject+".csv" };
      String [] types = { "text/comma-separated-values" };
      ipc.addItem("attachment_name", names);
      ipc.addItem("attachment_type", types);
      ipc.addItem("attachment0", csv.getBytes());
    } else {
      ipc.addItem("body", csv);
    }
    Registrar.sendMessage("Email" , ipc, null);
  }

  private void doNewFlipSheet(Event e) {
    setSheetNumber(e.data);
    // Kind of a crock because cancel will leave the new sheet there 
    // but with the default name.
    new Event(this, kShowSheetSettingsEvent).sendToTarget();
    sheetWindow.invalidateSheetNames();
  }

  //
  // Sheet number maintenance
  //
  int getSheetNumber() {
    return sheetNumber;
  }

  int getSheetUID() {
    return sheetUID;
  }

  int getSheetCount() {
    DataStore ds = getDataStore();
    if (ds == null) return 0;
    return ds.getRecordCount();
  }

  final SheetTable getSheetTable() {
    return sheetWindow.sheetTable;
  }

}
